#include "main.hpp"
#include "secrets.hpp"

#define WIRE Wire

#define SERVO_PIN 14     // D5
#define BUZZER_PIN 12    // D6
#define LEDSTRIPE_PIN 15 // D8

// buttons
#define TACTILE_BTN_LEFT 0  // D3
#define TACTILE_BTN_RIGHT 2 // D4
#define TOUCH_BTN_PIN 16    // D0

// RFID
#define PN532_IRQ 13   // D7 (not connected)
#define PN532_RESET 12 // D6 (not connected)
// Or use hardware Serial:
Adafruit_PN532 nfc(PN532_IRQ, PN532_RESET); // RFID with I2C

// OLED display
#define DISPLAY_PIN 4 // D2 (SDA)
auto display = Adafruit_SSD1306(128, 32, &WIRE);

// util classes
SmartStrip smartStrip;
SmartServo smartServo(SERVO_PIN);
TeaTimer teaTimer(&smartServo);
SmartRFID smartRFID(&nfc);
SmartDisplay smartDisplay(&display);
TeaNetworking teaNet;

SmartSound sound(BUZZER_PIN);

SmartButton buttonLeft(TACTILE_BTN_LEFT);
SmartButton buttonRight(TACTILE_BTN_RIGHT);

// networking
String shuttleServer = "https://ias-tea-axum.shuttleapp.rs";

TeaData currentTea("Teafault", "c0derfid", -1, 30);

enum State
{
  // lifecycle states
  WAIT_FOR_RFID_SCAN,
  PROCESS_RFID_SCAN,
  DISPLAY_SCANNED_TEA_CONFIG,
  STEEPING_IN_PROGRESS,
  REQUEST_FEEDBACK,

  // special states
  DISPLAY_ADD_NEW_TEA_CONFIG,

  // debugging states
  NETWORKING_DEBUG,
  SERVO_TUNING,
};
State state = State::WAIT_FOR_RFID_SCAN;

void setup()
{
  Serial.begin(9600);

  smartServo.moveServoToZero();

  smartStrip.init();

  smartStrip.setGreen(0);
  smartStrip.setGreen(7);
  smartStrip.setGreen(14);

  smartRFID.init();

  Serial.println("Initializing OLED...");
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C, false); // Address 0x3C for 128x32
  smartDisplay.printMsg("Tea is Great!");
  delay(500);

  smartDisplay.printMsg("waiting for internet connection");

  teaNet.init(ssid, password);

  smartDisplay.printMsg("Success!");
  delay(500);

  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(TOUCH_BTN_PIN, INPUT);

  buttonLeft.init();
  buttonRight.init();
}

void loop()
{

  switch (state)
  {
  case SERVO_TUNING:
  {
    // default min 1000, max 2000
    smartDisplay.printMsg("Write servo 0");
    smartServo.myservo.write(0);
    delay(2000);

    smartDisplay.printMsg("Write servo 180");
    smartServo.myservo.write(180);
    delay(2000);
    break;
  }
  case NETWORKING_DEBUG:
  {
    if (!teaNet.wifiConnected())
    {
      state = WAIT_FOR_RFID_SCAN;
      break;
    }
    JSONVar jsonData = teaNet.httpsGETRequest(shuttleServer + "/alltea");
    Serial.println("JSON:");
    Serial.println(jsonData[2]["teaname"]);
    delay(2000);

    String payload = teaNet.createTeaData("Unnamed Tea", "g8OhgB5", 90, 180);
    teaNet.httpsPOSTRequest("https://ias-tea-axum.shuttleapp.rs/addtea", payload);

    state = WAIT_FOR_RFID_SCAN;
    break;
  }
  case WAIT_FOR_RFID_SCAN:
  {
    smartDisplay.printWaitForRFIDScreen();

    // this blocks execution until a tag has been detected (5s timeout)
    String uid = smartRFID.readTag(5000);
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      state = PROCESS_RFID_SCAN;
    }
    break;
  }
  case PROCESS_RFID_SCAN:
  {

    display.println(currentTea.m_rfidCode);
    display.display();
    delay(500);

    display.println("connecting...");
    display.display();
    delay(250);

    JSONVar result = teaNet.httpsGETRequest(shuttleServer + "/teabyrfid/" + currentTea.m_rfidCode);
    Serial.print("Result:");
    Serial.println(result);

    // catch possible failure
    if (result == JSON.parse("{}"))
    {
      smartDisplay.printMsg("GET FAILED");
      delay(1000);
      break;
    }

    if (JSONVar::stringify(result) == "null")
    {
      Serial.print("TEA DOES NOT EXIST YET!");
      String payload = teaNet.createTeaData("Unnamed Tea", currentTea.m_rfidCode, 90, 180);
      result = teaNet.httpsPOSTRequest(shuttleServer + "/addtea", payload);
      // catch possible failure
      if (result == JSON.parse("{}"))
      {
        smartDisplay.printMsg("POST FAILED");
        delay(1000);
        break;
      }
      state = State::DISPLAY_ADD_NEW_TEA_CONFIG;
    }
    else
    {
      state = State::DISPLAY_SCANNED_TEA_CONFIG;
    }

    currentTea = TeaData(result["teaname"], currentTea.m_rfidCode, result["watertemp"], result["steepingseconds"]);
    break;
  }
  case DISPLAY_SCANNED_TEA_CONFIG:
  {

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (buttonLeft.isJustPressed())
    {
      currentTea.m_steepingSeconds -= 10;
    }
    if (buttonRight.isJustPressed())
    {
      currentTea.m_steepingSeconds += 10;
    }
    smartDisplay.printTeaConfigScreen(currentTea);

    if (isBtnPressed() && currentTea.m_steepingSeconds <= 0)
    {
      // rickroll
      Serial.println("Easter");
      sound.play(SmartSound::NeverGonna, 155);
      break;
    }

    // change state if start pressed
    if (isBtnPressed())
    {
      String payload = teaNet.createTeaData(currentTea);
      Serial.println(payload);
      smartDisplay.printMsg("sync...");
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/updatetea", payload);
      // catch possible failure
      if (res == JSON.parse("{}"))
      {
        smartDisplay.printMsg("POST FAILED");
        delay(1000);
        break;
      }
      teaTimer.beginSteeping(currentTea.m_steepingSeconds * 1000);
      state = State::STEEPING_IN_PROGRESS;
      smartStrip.reset();
    }
    break;
  }
  case STEEPING_IN_PROGRESS:
  {
    teaTimer.tick();
    if (teaTimer.isSteeping())
    {
      // abort if red button pressed
      if (buttonLeft.isJustPressed())
      {
        teaTimer.endSteeping();
        smartDisplay.printMsg("DONE");
        delay(1000);
        state = State::REQUEST_FEEDBACK;
        break;
      }

      auto r = teaTimer.remainingSteepTimeSeconds();
      smartDisplay.printTeaSteepingProgressScreen(currentTea, r);

      // update led strip
      smartStrip.progressBar(0, currentTea.m_steepingSeconds, r);
    }
    else
    {
      smartDisplay.printMsg("DONE");
      delay(1000);
      state = State::REQUEST_FEEDBACK;
    }
    break;
  }
  case REQUEST_FEEDBACK:
  {
    smartDisplay.printRequestFeedbackScreen(currentTea);

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (isBtnPressed())
    {
      // keep steeping time for next brew
      smartDisplay.printMsg("Feedback accepted (perfect)");
      delay(1000);
      state = State::WAIT_FOR_RFID_SCAN;
    }
    if (buttonLeft.isJustPressed())
    {
      // decrease steeping time for next brew
      smartDisplay.printMsg("Feedback accepted (-)");
      delay(1000);
      state = State::WAIT_FOR_RFID_SCAN;
    }
    if (buttonRight.isJustPressed())
    {
      // increase steeping time for next brew
      smartDisplay.printMsg("Feedback accepted (+)");
      delay(1000);
      state = State::WAIT_FOR_RFID_SCAN;
    }
    break;
  }
  case DISPLAY_ADD_NEW_TEA_CONFIG:
  {
    smartDisplay.printAddNewTeaConfigScreen(currentTea);

    String uid = smartRFID.readTag(50); // just enough time to detect tag
    if (uid != "no_tag_detected" && uid != "")
    {
      currentTea.m_rfidCode = uid;
      smartDisplay.resetDisplay();
      state = PROCESS_RFID_SCAN;
      break;
    }

    if (buttonLeft.isJustPressed())
    {
      currentTea.m_steepingSeconds -= 10;
    }
    if (buttonRight.isJustPressed())
    {
      currentTea.m_steepingSeconds += 10;
    }

    if (isBtnPressed())
    {
      smartDisplay.printMsg("Accepted");
      delay(500);
      smartDisplay.printMsg("sync...");
      JSONVar res = teaNet.httpsPOSTRequest(shuttleServer + "/updatetea", teaNet.createTeaData(currentTea));
      // catch possible failure
      if (res == JSON.parse("{}"))
      {
        smartDisplay.printMsg("POST FAILED");
        delay(1000);
        break;
      }
      state = State::DISPLAY_SCANNED_TEA_CONFIG;
    }

    break;
  }
  };
}

bool isBtnPressed()
{
  return digitalRead(TOUCH_BTN_PIN) == HIGH;
}